/* This file is part of Idest.
   Copyright (C) 2009-2011 Sergey Poznyakoff
 
   Idest is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
 
   Idest is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with Idest.  If not, see <http://www.gnu.org/licenses/>. */

#include "idest.h"

#ifdef GUILE_VERSION_NUMBER
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include <libguile.h>
#include <setjmp.h>

int guile_inited = 0;
int guile_debug = 1;
char *guile_script;
char *guile_function;
char **guile_argv;

SCM_GLOBAL_VARIABLE_INIT (sym_idest_main, "idest-main", SCM_EOL);
SCM_GLOBAL_VARIABLE_INIT(sym_idest_readonly, "idest-readonly", SCM_BOOL_F);

SCM_GLOBAL_SYMBOL(idest_text, "text");
SCM_GLOBAL_SYMBOL(idest_lang, "lang");
SCM_GLOBAL_SYMBOL(idest_class, "class");
SCM_GLOBAL_SYMBOL(idest_descr, "descr");



static SCM
eval_catch_body(void *list)
{
	SCM pair = (SCM)list;
	return scm_apply_0(SCM_CAR(pair), SCM_CDR(pair));
}

static SCM
eval_catch_handler(void *data, SCM tag, SCM throw_args)
{
	scm_handle_by_message_noexit("idest", tag, throw_args);
	longjmp(*(jmp_buf*)data, 1);
}

struct scheme_exec_data {
	SCM (*handler) (void *data);
	void *data;
};

static SCM
scheme_safe_exec_body(void *data)
{
	struct scheme_exec_data *ed = data;
	return ed->handler(ed->data);
}

static int
guile_safe_exec(SCM (*handler)(void *data), void *data, SCM *result)
{
	jmp_buf jmp_env;
	struct scheme_exec_data ed;
	SCM res;
	
	if (setjmp(jmp_env))
		return 1;
	ed.handler = handler;
	ed.data = data;
	res = scm_c_catch(SCM_BOOL_T,
			  scheme_safe_exec_body, (void*)&ed,
			  eval_catch_handler, &jmp_env,
			  NULL, NULL);
	if (result)
		*result = res;
	return 0;
}

struct load_closure {
	char *filename;
	int argc;
	char **argv;
};

static SCM
load_handler(void *data)
{
	struct load_closure *lp = data;
    
	scm_set_program_arguments(lp->argc, lp->argv, lp->filename);
	scm_primitive_load(scm_from_locale_string(lp->filename));
	return SCM_UNDEFINED;
}

static SCM
load_handler_path(void *data)
{
	struct load_closure *lp = data;
    
	scm_set_program_arguments(lp->argc, lp->argv, lp->filename);
	scm_primitive_load_path(scm_from_locale_string(lp->filename));
	return SCM_UNDEFINED;
}

static int
guile_load(char *filename, int use_path, char **argv)
{
	struct load_closure lc;
	if (argv) {
		lc.argc = -1;
		lc.argv = argv;
	} else {
		char *s_argv[2];
		s_argv[0] = filename;
		s_argv[1] = NULL;
		lc.argc = 1;
		lc.argv = s_argv;
	}
	lc.filename = filename;
	return guile_safe_exec(use_path ? load_handler_path : load_handler,
			       &lc, NULL);
}

static SCM
stringlinst_to_scm(union id3_field *field, int genre)
{
	unsigned i, nstrings = id3_field_getnstrings(field);
	struct obstack stk;
	char *str;
	SCM ret;
	
	obstack_init(&stk);
	for (i = 0; i < nstrings; i++) {
		id3_ucs4_t const *ucs4;
		
		ucs4 = id3_field_getstrings(field, i);
		if (!ucs4)
			continue;
		if (genre)
			ucs4 = id3_genre_name(ucs4);
		str = idest_ucs4_cvt(ucs4);
		obstack_grow(&stk, str, strlen(str));
		free(str);
	}
	obstack_1grow(&stk, 0);
	str = obstack_finish(&stk);
	ret = scm_from_locale_string(str);
	obstack_free(&stk, NULL);
	return ret;
}

static SCM
field_to_scm(union id3_field *field, int genre)
{
	id3_ucs4_t const *ucs4;
	char *str;
	SCM ret = SCM_EOL;
	
	switch (id3_field_type(field)) {
	case ID3_FIELD_TYPE_TEXTENCODING:
	case ID3_FIELD_TYPE_INT8:
	case ID3_FIELD_TYPE_INT16:
	case ID3_FIELD_TYPE_INT24:
	case ID3_FIELD_TYPE_INT32:
	case ID3_FIELD_TYPE_INT32PLUS:
		ret = scm_from_long(field->number.value);
		break;

	case ID3_FIELD_TYPE_LATIN1:
	case ID3_FIELD_TYPE_LATIN1FULL:
		/* FIXME */
		ret = scm_from_locale_string(id3_field_getlatin1(field));
		break;
		
	case ID3_FIELD_TYPE_LATIN1LIST:
		/* FIXME */
		break;

	case ID3_FIELD_TYPE_STRING:
		ucs4 = id3_field_getstring(field);;
		if (ucs4) {
			str = idest_ucs4_cvt(ucs4);
			ret = scm_from_locale_string(str);
			free(str);
		} else
			ret = SCM_BOOL_F;
		break;
		
	case ID3_FIELD_TYPE_STRINGFULL:
		ucs4 = id3_field_getfullstring(field);
		if (ucs4) {
			str = idest_ucs4_cvt(ucs4);
			ret = scm_from_locale_string(str);
			free(str);
		} else
			ret = SCM_BOOL_F;
		break;
		
	case ID3_FIELD_TYPE_STRINGLIST:
		ret = stringlinst_to_scm(field, genre);
		break;
		
	case ID3_FIELD_TYPE_LANGUAGE:
	case ID3_FIELD_TYPE_FRAMEID:
	case ID3_FIELD_TYPE_DATE:
		ret = scm_from_locale_string(field->immediate.value);
		break;
		
	case ID3_FIELD_TYPE_BINARYDATA:
		/* FIXME */
		ret = SCM_EOL;
	}
	return ret;
}

static SCM
frame_to_scm(struct id3_frame *frame)
{
	SCM val;
	union id3_field *field;

	switch (frame_id(frame->id)) {
	case item_comment:
		field = id3_frame_field(frame, 3);
		val = scm_list_1(
			scm_cons(idest_text,
				 scm_from_locale_string(
					 field_to_string(field, 0))));
					  
		field = id3_frame_field(frame, 1);
		if (!field)
			break;
		val = scm_cons(
			scm_cons(idest_lang,
				 scm_from_locale_string(
					 field_to_string(field, 0))),
			val);
		field = id3_frame_field(frame, 2);
		if (!field)
			break;
		val = scm_cons(
			scm_cons(idest_class,
				 scm_from_locale_string(
					 field_to_string(field, 0))),
			val);
		break;
			
	case item_genre:
		val = scm_list_1(
			scm_cons(idest_text,
				 field_to_scm(id3_frame_field(frame, 1), 1)));
		break;
		
	default:
		val = scm_list_1(
			scm_cons(idest_text,
				 field_to_scm(id3_frame_field(frame, 1), 0)));
		break;
	}
	return scm_cons(scm_from_locale_string(frame->id),
			scm_cons(
			   scm_cons(idest_descr,
 			      scm_from_locale_string(frame->description)),
			   val));
}

static SCM
tag_to_scm(struct id3_tag *tag)
{
	struct id3_frame *frame;
	unsigned i;
	SCM scm_first = SCM_EOL, scm_last;

	for (i = 0; (frame = id3_tag_findframe(tag, NULL, i)); i++) {
		SCM new;
		SCM fcell = frame_to_scm(frame);
		
		if (fcell == SCM_EOL)
			continue;
		new = scm_cons(fcell, SCM_EOL);
		if (scm_first == SCM_EOL) 
			scm_last = scm_first = new;
		else {
			SCM_SETCDR(scm_last, new);
			scm_last = new;
		}
	}
	return scm_first;
}

static int
scm_to_tag(SCM scm, struct id3_tag *tag)
{
	int modified = 0;
	
	for (; !scm_is_null(scm) && scm_is_pair(scm); scm = SCM_CDR(scm)) {
		struct id3_frame *frame;
		struct ed_item itm;
		char *id;
		SCM x, text;
		SCM elt = SCM_CAR(scm);
		if (!scm_is_pair(elt))
			scm_misc_error(NULL,
				       "Wrong element type: ~S",
				       scm_list_1(elt));
		
		x = SCM_CAR(elt);
		if (!scm_is_string(x))
			scm_misc_error(NULL,
				       "Wrong car type: ~S",
				       scm_list_1(elt));
		id = scm_to_locale_string(x);
		
		memset(&itm, 0, sizeof(itm));

		x = SCM_CDR(elt);

		if (scm_is_string(x)) {
			itm.v.value = scm_to_locale_string(x);
		} else if (scm_is_pair(x) &&
			   (text = scm_assoc_ref(x, idest_text)) != SCM_BOOL_F) {
			itm.v.value = scm_to_locale_string(text);
			text = scm_assoc_ref(x, idest_lang);
			if (text != SCM_BOOL_F)
				itm.lang = scm_to_locale_string(text);
			text = scm_assoc_ref(x, idest_class);
			if (text != SCM_BOOL_F)
				itm.class = scm_to_locale_string(text);
		} else
			scm_misc_error(NULL,
				       "Wrong cdr type: ~S",
				       scm_list_1(elt));

		frame = id3_frame_new(id);
		if (id3_tag_attachframe(tag, frame))
			error(1, 0, "cannot attach new frame");
		set_frame_value(frame, &itm);
		modified |= 1;
		free(id);
		free(itm.v.value);
		free(itm.class);
		free(itm.lang);
	}
	return modified;
}

SCM
guile_apply_main(const char *file, struct id3_tag *tag)
{
	jmp_buf jmp_env;
	SCM cell;

	if (setjmp(jmp_env)) 
		error(1, 0, "idest-main failed");

	cell = scm_cons(SCM_VARIABLE_REF(sym_idest_main),
			scm_list_2(scm_from_locale_string(file),
				   tag_to_scm(tag)));
	return scm_c_catch(SCM_BOOL_T,
			   eval_catch_body, cell,
			   eval_catch_handler, &jmp_env,
			   NULL, NULL);
}

int
guile_transform(const char *file, struct id3_tag *tag)
{
	SCM result;

	if (!guile_inited)
		return 0;
    
	result = guile_apply_main(file, tag);

	if (scm_is_pair(result)) {
		/* Remove all existing tags */
		id3_tag_clearframes(tag);
		/* Replace them with the new ones */
		return scm_to_tag(result, tag);
	}
	return 0;
}

int
guile_list(const char *file, struct id3_tag *tag)
{
	if (!guile_inited)
		return 0;
    
	guile_apply_main(file, tag);
	return 1;
}

static void
load_path_prepend(const char *dir)
{
	SCM scm, path_scm;
	SCM *pscm;

	path_scm = SCM_VARIABLE_REF(scm_c_lookup("%load-path"));
	for (scm = path_scm; !scm_is_null(scm); scm = SCM_CDR(scm)) {
		SCM val = SCM_CAR(scm);
		if (scm_is_string(val)) {
			char *s = scm_to_locale_string(val);
			int rc = strcmp(s, dir);
			free(s);
			if (rc == 0)
				return;
		}
	}

	pscm = SCM_VARIABLE_LOC(scm_c_lookup("%load-path"));
	*pscm = scm_cons(scm_from_locale_string(dir), path_scm);
}

static int
try_file(const char *dir, const char *file_name)
{
	char *ptr;
	size_t len = strlen(dir);
	if (len > 0 && dir[len - 1] == '/')
		len--;
	ptr = xmalloc(len + 1 + strlen(file_name) + 1);
	memcpy(ptr, dir, len);
	ptr[len++] = '/';
	strcpy(ptr + len, file_name);

	if (access(ptr, R_OK) == 0) {
		if (guile_load(ptr, 0, NULL))
			error(1, 0, "cannot load startup script %s", ptr);
		return 0;
	}
	free(ptr);
	return 1;
}

static void
load_startup_file()
{
	const char init_name[] = ".idest.scm";
	const char *dir;

	load_path_prepend(GUILE_SITE);
	load_path_prepend(PKG_SITE);
	load_path_prepend(".");
	
	if (access(init_name, R_OK) == 0) {
		if (guile_load((char*)init_name, 0, NULL) == 0)
			return;
		error(1, 0, "cannot load startup script %s", init_name);
	}
	
	dir = getenv("HOME");
	if (dir && try_file(dir, init_name) == 0)
		return;
	try_file(PKG_SITE, "idest.scm");
}

void
guile_init(int *pargc, char ***pargv)
{
	SCM readonly, proc, args;
	int argc, i;
	char **argv;
	
	if (!guile_script)
		return;

	scm_init_guile();
	scm_load_goops();
#include "guile.x"
	
	if (guile_debug) {
#ifdef GUILE_DEBUG_MACROS
		SCM_DEVAL_P = 1;
		SCM_BACKTRACE_P = 1;
		SCM_RECORD_POSITIONS_P = 1;
		SCM_RESET_DEBUG_MODE;
#endif
	}

	load_startup_file();

	guile_load(guile_script, !strchr(guile_script, '/'), guile_argv);

	/* Read command line arguments */
	args = scm_program_arguments();
	argc = scm_to_int(scm_length(args));
	/* Note: argv[0] is the script name. We will skip it. */
	argv = xcalloc(argc, sizeof(argv[0]));
	argc--;
	for (i = 0, args = SCM_CDR(args); i < argc;
	     i++, args = SCM_CDR(args)) {
		SCM sarg = SCM_CAR(args);
		argv[i] = scm_to_locale_string(sarg);
	}
	argv[i] = NULL;
	*pargc = argc;
	*pargv = argv;
	
	proc = SCM_VARIABLE_REF(sym_idest_main);
	if (proc == SCM_EOL) {
		if (guile_function) {
			proc = SCM_VARIABLE_REF(scm_c_lookup(guile_function));
			SCM_VARIABLE_SET(sym_idest_main, proc);
		} else
			error(1, 0, "idest-main not defined");
	}
				
	if (scm_procedure_p(proc) != SCM_BOOL_T) 
		error(1, 0,
		      "idest-main is not a procedure object");

	readonly = SCM_VARIABLE_REF(sym_idest_readonly);
	if (readonly == SCM_BOOL_F) 
		mode = MODE_MOD;
	else if (readonly == SCM_BOOL_T)
		mode = MODE_QUERY;
	else
		error(1, 0,
		      "script %s set non-boolean value of idest-readonly",
		      guile_script);
	guile_inited = 1;
}
#else

int
guile_transform(const char *file, struct id3_tag *tag)
{
	return 0;
}

int
guile_list(const char *file, struct id3_tag *tag)
{
	return 0;
}

void
guile_init(int *pargc, char **pargv)
{
}

#endif
